"""Several constants which are used for the loudness converter module.

"""

AUDITORY_THRESHOLD_AT_1KHZ = 0.00002
"""Roughly the sound of a mosquito flying 3 m away
(see https://en.wikipedia.org/wiki/Sound_pressure)."""
